# = dev-utils/debug/irb.rb
#
# DevUtils::Debug.breakpoint and some aliases are implemented here.  Do not load this file
# directly.
#

require 'irb'
require 'extensions/binding'

module DevUtils::Debug

    #
    # Method:: *breakpoint*
    # Aliases:: *break_point*, *goirb*
    #
    # === Description
    # 
    # This will pop up an interactive ruby session from whereever it is called in a Ruby
    # application. In IRB you can examine the environment of the break point, peeking
    # and poking local variables, calling methods, viewing the stack (with +caller+), etc.
    #
    # This is like setting breakpoints in a debugger, except that you are running the program
    # normally (debuggers tend to run programs more slowly).  Debuggers are generally more
    # flexible, but this is a good lightweight solution for many cases.  You <em>can not</em>
    # step through the code with this technique.  But you can, of course, set multiple
    # breakpoints.  And you can make breakpoints conditional.
    #
    # You can force a breakpoint to return a certain value.  This is typically only useful if
    # the breakpoint is the last value in a method, as this will cause the method to return a
    # different value than normal.  This is demonstrated in the example below.
    #
    # You can also give names to break points which will be used in the message that is
    # displayed upon execution of them.  This helps to differentiate them at runtime if you
    # have set several breakpoints in the code.
    #
    # === Parameters
    #
    # _name_:: A String to identify the breakpoint, giving a more informative message.
    # _context_:: Any object; IRB will use this as its context.  The default is the current
    #             scope's binding, which is nearly always what you will want.
    # _block_:: Will be executed when the breakpoint returns normally.  Bypassed if you
    #           <tt>throw :debug_return, </tt><em>value</em> from IRB.  Unless you are
    #           _planning_ to use the <tt>debug_return</tt> feature for a given breakpoint,
    #           you don't need to worry about the block.
    #
    # === Typical Invocation
    #
    #   breakpoint                          # Plain message.
    #   breakpoint "Person#name"            # More informative message.
    #   breakpoint { normal_return_value }
    #
    # === Example
    #
    # Here's a sample of how breakpoints should be placed:
    #
    #   require 'dev-utils/debug'
    #
    #   class Person
    #     def initialize(name, age)
    #       @name, @age = name, age
    #       breakpoint "Person#initialize"
    #     end
    #
    #     attr_reader :age
    #     def name
    #       breakpoint "Person#name" { @name }
    #     end
    #   end
    #
    #   person = Person.new("Random Person", 23)
    #   puts "Name: #{person.name}"
    #
    # And here is a sample debug session:
    #
    #   Executing break point "Person#initialize" at file.rb:4 in `initialize'
    #   irb(#<Person:0x292fbe8>):001:0> <b>local_variables</b>
    #   => ["name", "age", "_", "__"]
    #   irb(#<Person:0x292fbe8>):002:0> [name, age]
    #   => ["Random Person", 23]
    #   irb(#<Person:0x292fbe8>):003:0> [@name, @age]
    #   => ["Random Person", 23]
    #   irb(#<Person:0x292fbe8>):004:0> self
    #   => #<Person:0x292fbe8 @age=23, @name="Random Person">
    #   irb(#<Person:0x292fbe8>):005:0> @age += 1; self
    #   => #<Person:0x292fbe8 @age=24, @name="Random Person">
    #   irb(#<Person:0x292fbe8>):006:0> exit
    #   Executing break point "Person#name" at file.rb:9 in `name'
    #   irb(#<Person:0x292fbe8>):001:0> throw(:debug_return, "Overriden name")
    #   Name: Overriden name
    #
    # This example is explored more thoroughly at
    # http://dev-utils.rubyforge.org/DebuggingAids.html.
    #
    # === Credits
    #
    # Joel VanderWerf and Florian Gross have contributed the code and documentation for this
    # method.
    #
  def breakpoint(name = nil, context = nil, &block)
    file, line, method = *caller.first.match(/^(.+?):(\d+)(?::in `(.*?)')?/).captures
    message = "Executing breakpoint #{name.inspect if name} at #{file}:#{line}"
    message << " in '#{method}'" if method

    body = lambda do |_context|
      puts message
      catch(:debug_return) do |value|
        IRB.start_session(IRB::WorkSpace.new(_context))
        block.call if block        
      end
    end

      # Run IRB with the given context, if it _is_ given.
    return body.call(context) if context
      # Otherwise, run IRB with the parent scope's binding, giving access to
      # the local variables of the method that called method.
    Binding.of_caller do |binding_context|
      body.call(binding_context)
    end
  end

  alias :break_point :breakpoint
  alias :goirb       :breakpoint
  
  alias lv local_variables
  alias iv instance_variables
end

# :stopdoc:

module ::IRB
   
   class << self; remove_method :parse_opts; end
   def IRB.parse_opts
     # Don't touch ARGV, which belongs to the app which called this module.
   end

   def IRB.start_session(object)
     unless $irb
       IRB.setup nil
       ## maybe set some opts here, as in parse_opts in irb/init.rb?
     end

     @CONF[:PROMPT_MODE] = :SIMPLE

     workspace = WorkSpace.new(object)

     if @CONF[:SCRIPT] ## normally, set by parse_opts
       $irb = Irb.new(workspace, @CONF[:SCRIPT])
     else
       $irb = Irb.new(workspace)
     end

     @CONF[:IRB_RC].call($irb.context) if @CONF[:IRB_RC]
     @CONF[:MAIN_CONTEXT] = $irb.context

     trap 'INT' do
       $irb.signal_handle
     end

     custom_configuration if defined?(IRB.custom_configuration)

     catch :IRB_EXIT do
       $irb.eval_input
     end

     puts

     ## might want to reset your app's interrupt handler here
   end
end

class Object
   include IRB::ExtendCommandBundle # so that Marshal.dump works
end

# :startdoc:
